# -*- coding: utf-8 -*-
first_name = '' # TODO
last_name = '' # TODO
hand_in = False # TODO: set to True and run again before hand_in. Hand_in also the output* file

import sys
import os
if hand_in:
    filename = os.path.basename(__file__).split('.')[0]
    output = open(f'output_{filename}_{last_name}_{first_name}.txt', 'w')
    sys.stdout = output
    print(f"{filename}\n {last_name}, {first_name}")

from numpy import zeros, fft, sqrt, pi, real, linspace, sin, cos, where, linalg, arange, finfo
from pylab import semilogy, show, plot, legend, xlabel, ylabel, savefig, grid, title, subplots

# set up symbolic functions
import sympy as sy
x_  = sy.symbols('x') # define symbols
fn = sy.exp(-1/sy.sin(2*pi*x_)**2) # smooth but not analytic
fa= 1/(1+sy.sin(2*pi*x_)**2) # analytic in a strip in complex plane
fb = sy.sin(2*pi*10*x_) # band limted

# make them and their derivatives usable with numpy
syfuncs = [fn, fa, fb]
def sy2np(func):
    f1 = sy.lambdify((x_), func, "numpy")
    dfunc = sy.diff(func, x_, 1) # its first derivative
    dfunc = sy.simplify(dfunc)
    df1 = sy.lambdify((x_), dfunc, "numpy")
    return f1, df1    

# sympy has dificulties with abs so here is the plain implementation
def absin3(x):
    return abs(sin(2*pi*x))**3
def dabsin3(x):
    return 3*2*pi*cos(2*pi*x)*sin(2*pi*x)*abs(sin(2*pi*x))

# a list of all functions
LofF =  list(map(sy2np,syfuncs))
LofF.append((absin3, dabsin3))

# evaluate a trigonometric interpolation polynomial
def evaliptrig(y,N): # y = values, N=number of points to evaluate
    n = len(y)
    if (n%2) == 0:
       # TODO: replace the following line by the right code
       v = zeros(N, dtype=complex)      
       return v
    else: print('odd length')

# evaluate the derivative via trigonometric interpolation polynomial
def evalDitrig(y,N): # y = values, N=number of points to evaluate
    n = len(y)
    if (n%2) == 0:
       # TODO: replace the following line by the right code
       v = zeros(N, dtype=complex) 
       return v
    else: print('odd length')

#
# we have to set the left corner slightly larger than 0
# because the symbolic derivative of fn needs a lim at 0
a = finfo(float).eps 
b = 1. 
# evaluate functions or derivatives and convergence rates
def efc(f, df='None',N=2**12, lev=7):
    tt = linspace(a,b,N, endpoint=False) # reference points to evaluate
    if df == 'None':
        fv = f(tt) # evaluate into reference points
        worker = evaliptrig # evaluate function
    else:
        fv = df(tt) # evaluate into reference points
        worker = evalDitrig # evaluate derivative
    n = 2
    vs = []
    vlinf = []; vl2 = []; vn = []; vlinfloc = []
    while n < 1+2**lev:
        t = linspace(0,1,n, endpoint=False) # interpolation points
        y = f(t) # evaluate the function there
        # evaluate the trig. interpolant in N points
        v = real(worker(y,N)) # only real functions
        vs += [v]
        d = abs(v-fv); linf = d.max() 
        l2 = linalg.norm(d)/sqrt(N)
        linfloc = d[5*N//16:7*N//16].max()
        vlinf += [linf]; vl2 += [l2]
        vlinfloc +=[linfloc]
        vn += [n]
        n *= 2
    return (tt, fv, vs, vn, vl2, vlinf, vlinfloc)


# run for convergence plots of functions
fig, ax = subplots(2, figsize=(12, 12))
titles = ['smooth non-analytic n(x)', 'analytic a(x)','band limited b(x)','C^3 v(x)']
fname = ['u(x)', 'a(x)', 'b(x)', 'v(x)']
marker = ['x', '+', 'o', '.'] 
k = 0
for fus in LofF:
    print('now: ',fus[0].__name__, fus[0].__doc__,'----\n')
    tt, fv, vs, vn, vl2, vlinf, vlinfloc = efc(fus[0], lev=8)
    ax[0].semilogy(vn, vl2, '-', marker=marker[k], label=fname[k])
    tt, fv, vs, vn, vl2, vlinf, vlinfloc =  efc(*fus, lev=10)
    print(vl2)
    ax[1].semilogy(vn, vl2, '-', marker=marker[k], label=fname[k])
    k += 1
print('oyt of cycle')    
ax[0].set_title('Functions')
ax[0].set_ylabel('error'); ax[0].set_xlabel('n')
ax[0].grid()
ax[1].set_title('Derivatives')
ax[1].set_ylabel('error'); ax[1].set_xlabel('n')
ax[1].grid()
ax[0].legend(loc='center right', bbox_to_anchor=(1.1, 0.5))
ax[1].legend(loc='center right', bbox_to_anchor=(1.1, 0.5))
savefig('Err.png')
show()

